/* modelBuildingTools - Software to create texture and shape models
   for Active Models

    Jesus Nuevo-Chiquero, University of Alcal, Spain. 2010
    jesus.nuevochiquero@gmail.com

    This file is part of 'modelBuildingTools'.

    modelBuildingTools is free software: you can redistribute it
    and/or modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    modelBuildingTools is distributed in the hope that it will be
    useful, but WITHOUT ANY WARRANTY; without even the implied
    warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with modelBuildingTools.  If not, see
    <http://www.gnu.org/licenses/>.

*/
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <stddef.h>
#include <dirent.h>

#include "basic_utils.hh"

using namespace cv;

// A helper function to simplify the main part.
//template<class T>
std::ostream& operator<<(std::ostream& os,  Mat& v)
{

    if(v.type() == CV_64F){
	for(uint ai=0; ai<v.rows;ai++){
	    Mat row = v.row(ai);
	    MatIterator_<double> b = row.begin<double>();
	    MatIterator_<double> e = row.end<double>();
	    copy(b,e, std::ostream_iterator<double>(os, " ")); 
	    os<<std::endl;
	}
    }
    if(v.type() == CV_32F){
	for(uint ai=0; ai<v.rows;ai++){
	    Mat row = v.row(ai);
	    MatIterator_<float> b = row.begin<float>();
	    MatIterator_<float> e = row.end<float>();
	    copy(b,e, std::ostream_iterator<float>(os, " ")); 
	    os<<std::endl;
	}
    }

    return os;
}


void print_arr(const CvArr *arr, bool transpose)
{

    CvMat *mat, head;
    mat = cvGetMat(arr, &head);
  
    if(!transpose){
	for(int ai=0; ai<mat->rows; ai++){
	    for(int bi=0; bi<mat->cols; bi++){
		std::cout << cvGetReal2D(mat, ai, bi) << " ";
	    }
	    std::cout << std::endl;
	}
    }
    else{
	for(int bi=0; bi<mat->cols; bi++){
	    for(int ai=0; ai<mat->rows; ai++){
		std::cout << cvGetReal2D(mat, ai, bi) << " ";
	    }
	    std::cout << std::endl;
	}

    }

}


static int select_imgs(const struct dirent *eimg )
{
    
    std::string file = eimg->d_name, exten;
    
    if(eimg->d_type != DT_REG)
	return 0;

    size_t pos = file.find_last_of(".");
    if(pos == std::string::npos)
	return 0;
    exten = file.substr(pos+1, file.size());
    if(exten == "pgm" || exten == "jpg" || exten == "jpeg"){
	return 1;
    }

    return 0;
}

static int select_bioidfiles(const struct dirent *emark)
{
    
    std::string file = emark->d_name, exten;
    
    if(emark->d_type != DT_REG)
	return 0;

    size_t pos = file.find_last_of(".");
    if(pos == std::string::npos)
	return 0;
    exten = file.substr(pos+1, file.size());
    if(exten == "pts"){
	return 1;
    }

    return 0;
}

static int select_xmlfiles(const struct dirent *emark)
{
    
    std::string file = emark->d_name, exten;
    
    if(emark->d_type != DT_REG)
	return 0;

    size_t pos = file.find_last_of(".");
    if(pos == std::string::npos)
	return 0;
    exten = file.substr(pos+1, file.size());
    if(exten == "xml"){
	return 1;
    }

    return 0;
}



int read_image_mark_files(std::vector<std::string> inputs, bool use_bioid, 
			  std::vector<std::string>&imgfiles, 
			  std::vector<std::string>&marks)
{

    if(inputs.size()==2){
	struct stat buf1, buf2;
	struct dirent **eimg, **emark;
	std::string &imgdir = inputs.at(0), &markdir = inputs.at(1);
	
	stat(imgdir.c_str(), &buf1);
	stat(markdir.c_str(), &buf2);
	
	if(S_ISDIR(buf1.st_mode) ^ S_ISDIR(buf2.st_mode)){
	    std::cerr << "Bad parameters, unable to read files" << std::endl;
	    return -1;
	}
	else if(S_ISDIR(buf1.st_mode) && S_ISDIR(buf2.st_mode)){
	    //read the folders
	    
	    int nimgs = scandir(imgdir.c_str(), &eimg, select_imgs, alphasort);
	    int nmarks;
	    if(use_bioid)
		nmarks = scandir(markdir.c_str(), &emark, select_bioidfiles, alphasort);
	    else
		nmarks = scandir(markdir.c_str(), &emark, select_xmlfiles, alphasort);
	    
	    
	    for(int ai=0; ai<MIN(nimgs, nmarks);ai++){
		std::string imgfile, markfile;
		imgfile = imgdir;
		imgfile.append("/");
		imgfile.append(eimg[ai]->d_name);
		markfile = markdir;
		markfile.append("/");
		markfile.append(emark[ai]->d_name);
		
		imgfiles.push_back(imgfile);
		marks.push_back(markfile);
	    }
	 }
    }
    else{
	if(inputs.size()%2){
	    std::cerr << "Bad number of parameters: there must be the same number of images and files with marks"<<std::endl;
	    return -1;
	}
	for(uint ai=0; ai<inputs.size()/2;ai++){
	    imgfiles.push_back(inputs.at(ai));
	    marks.push_back(inputs.at(2*ai));	    
	}

    }

    return imgfiles.size();
}
