//Creation of a texture orthonormal vector base using PCA from a set
//of samples.

/* modelBuildingTools - Software to create texture and shape models
   for Active Models

    Jesus Nuevo-Chiquero, University of Alcal, Spain. 2010
    jesus.nuevochiquero@gmail.com

    This file is part of 'modelBuildingTools'.

    modelBuildingTools is free software: you can redistribute it
    and/or modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    modelBuildingTools is distributed in the hope that it will be
    useful, but WITHOUT ANY WARRANTY; without even the implied
    warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with modelBuildingTools.  If not, see
    <http://www.gnu.org/licenses/>.

*/
#include <stdio.h>
#include <time.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <stddef.h>
#include <dirent.h>

#include <vector>
#include <stdlib.h>
#include <algorithm>
#include <iostream>
#include <fstream>
#include <iomanip>

#include <stdexcept>

#include <getopt.h>

#include <cxcore.h>
#include <cv.h>
#include <highgui.h>

#include <boost/smart_ptr.hpp>
#include <boost/program_options.hpp>

#include "basic_utils.hh"

#include "config.h"

using namespace cv;
namespace po = boost::program_options;

const char *const svn_str = "$Id: build_app_base.cpp 701 2009-12-22 16:36:01Z jnuevo $";
const char *const svn_rev = "$Rev: 944 $";

#define DEFAULT_KEEP_PCA 0.95
#define DEFAULT_MIN_STD_VALUE 1e-2

typedef struct cliparams_t{
    std::string output;
    std::string inputs; //folder where the samples are stored
    bool per_patch;    
    bool normalize;

    std::string discard;
    double keep_pca;    
    //    unsigned int max_patches;   
    int max_components;
}cliparams_t;


int parse_cli(int argc, char *argv[], cliparams_t &clarg)
{

    std::vector<int> size;

  try{

	po::options_description generic("Generic options");
	generic.add_options()
	    ("version,v", "print version string")
	    ("help,h", "produce help message")    
	    ;

	po::options_description config("Configuration");
        config.add_options()
	    ("keep,k", po::value<double>(&clarg.keep_pca),
	     "Representation percentage to be kept when performing PCA. Default 0.95")
	    ("max_components,m", po::value<int>(&clarg.max_components)->default_value(0),
	     "Maximum number of vectors, regardless if they cover the representation percentage (0 = no limit)")
	    ("discard,d", po::value<std::string>(&clarg.discard)->default_value(""),
	     "A file containing a 1D CvMat, with points to discard from the samples")
	    ("per-patch,p", "Build the vector base independently for each patch")
	    ("normalize,n", "Normalize patches to mean 0 and variance 1 before applying PCA")
	    ;

	po::options_description output("Output");
	output.add_options()
	    ("output,o", po::value<std::string>(&clarg.output), "Output XML file.")
	    ;
	
	// Hidden options, will be allowed both on command line and
        // in config file, but will not be shown to the user.
        po::options_description hidden("Hidden options");
        hidden.add_options()
            ("input-files", po::value< std::string >(&clarg.inputs), "input files")
            ;
	
  	po::positional_options_description p;
        p.add("input-files", 1);

	po::options_description visible;
        visible.add(generic).add(config).add(output);
	

	po::options_description cmdline_options;
        cmdline_options.add(generic).add(config).add(hidden).add(output);

        po::variables_map vm;
        store(po::command_line_parser(argc, argv).
              options(cmdline_options).positional(p).run(), vm);
	
        notify(vm);
	
	if(vm.count("version")){
	    std::cout << "Version: "<<modelBuildingTools_VERSION_MAJOR<<"."
		      <<modelBuildingTools_VERSION_MINOR<<"."
		      <<modelBuildingTools_VERSION_PATCH<<std::endl;
	    std::cout << "SVN rev. : " <<svn_rev << std::endl;
	    return 1;
	}

	if (vm.count("help")) {
	    std::cout << "Usage: build_app_base [options] <folder>"<<std::endl;
	    std::cout << visible << std::endl;
	    std::cout << "<folder> contains a series of files with the extracted patches, as\
extracted with extract_patches" << std::endl;
            return 1;
        }

	if(!vm.count("input-files")){
	    std::cout << "No input files"<<std::endl;
	    return 1;
	}

	if(!vm.count("keep")){
	    clarg.keep_pca = DEFAULT_KEEP_PCA;
	}

	if(!vm.count("output")){
	    for(uint ai=0; ai<1e5;ai++){
		char file[128];
		sprintf(file, "app_base_ouput_%05d.xml", ai);

		if(access (file, F_OK)){
		    clarg.output = file;
		    break;
		}
	    }
	    if(!clarg.output.length()){
		std::cerr << "Output file could not be created, too many files in current folder"<<std::endl;
		exit(1);
	    }		
	}

	if(vm.count("normalize")){
	    clarg.normalize = true;
	}
	else
	    clarg.normalize = false;

	if(vm.count("per-patch")){
	    clarg.per_patch = true;	   
	}
	else
	    clarg.per_patch = false;
	
    }
    catch(std::exception& e)
	{
	    std::cout << e.what() << "\n";
	    return 1;
	}    
    




    return 0;
}


int select_xmlfiles(const struct dirent *emark)
{
    
    std::string file = emark->d_name, exten;
    
    if(emark->d_type != DT_REG)
	return 0;

    size_t pos = file.find_last_of(".");
    if(pos == std::string::npos)
	return 0;
    exten = file.substr(pos+1, file.size());
    if(exten == "xml"){
	return 1;
    }

    return 0;
}


double normalizar(const Mat &in, Mat &m)
{

    //OpenCV's GaussianBlur will take values outside the matrix if it is part of a
    //bigger image. This is can not be disabled, it's a bug.


    //blurring the patches a bit reduces the number of eigenvectors roughly by half
    GaussianBlur(in, m, Size(3,3), 1,1);
  
    Scalar me, st;

    meanStdDev(m, me, st);
    m -= me;

    if(st.val[0] > DEFAULT_MIN_STD_VALUE)
 	m = m / st.val[0];
    
#ifdef DEBUG
    else
	std::cerr << "Low std value!: "<< st.val[0] << std::endl;
#endif			    
    
    return st.val[0];

}


Mat loadSamples(const std::string& folder, std::vector<std::vector<Mat> >& samples, 
			 const Mat& discard, bool complete_only = false)
{

    //the samples are named as "sample_%05d.xml"
    struct dirent **edata;
    int nfiles = scandir(folder.c_str(), &edata, select_xmlfiles, alphasort);
    std::vector<std::string> files;
    
    Mat count; //counter of samples (per patch)
         
    static int wait_time = 0;

    for(uint ai=0; ai<nfiles;ai++){
	std::vector<Mat> smpl;
  
	FileStorage fs(folder + "/" + edata[ai]->d_name, FileStorage::READ);

	if(fs.isOpened()){
	    int npatches, nOK = 0;
	    Size sz;
	    FileNode node = fs["patch_size"];
	    node[0] >> sz.width;
	    node[1] >> sz.height;
	    
	    fs["total_patches"] >> npatches;
	    fs["patches_written"] >> nOK;
	    

	    if(nOK==0){
		continue;
	    }
	    if(complete_only && (nOK != npatches))
		continue;

	    if(count.empty()){
		//		count.create(1, npatches, CV_32U);
		count.create(npatches,1,CV_32S);
		count = Scalar(0);

	    }


#ifdef DEBUG
	    namedWindow("baseAPP_DBG", CV_WINDOW_AUTOSIZE);
	    Mat dbg = Mat::zeros(sz.height*npatches, 2*sz.width, CV_32F);
	    
#endif

	    for(int n=0; n<npatches; n++){
		Mat m;
		char name[64];
		sprintf(name, "patch_%03d",n);
		if(!fs[name].empty()){
		    read(fs[name], m);
		    if(m.data){
			count.at<int>(n,0)++;

#ifdef DEBUG
			Mat mt = dbg(Range(n*sz.height, (n+1)*sz.height),Range(0, sz.width));
			m.copyTo(mt);
					
			//plot the normalized version on the side
			mt = dbg(Range(n*sz.height, (n+1)*sz.height),Range(sz.width, 2*sz.width));
			m.copyTo(mt);
			normalizar(m,mt);
		
#endif

		    }
		}
		


		smpl.push_back(m);//include it anyway		    		    	    
	    }
#ifdef DEBUG
		imshow("baseAPP_DBG", dbg);
		unsigned char key = static_cast<unsigned char> (waitKey(wait_time));
		if(key == '0' || key == ' '){
		    wait_time = 0;
		}		
		else if(key == '1'){
		    wait_time = 2;
		}
		else if(key == '2'){
		    wait_time = 10;
		}
		else if(key == '3'){
		    wait_time = 100;
		}
		else if(key == 'd'){
		    Mat d = dbg(Rect(sz.width,0,10,10));
		    std::cout << d << std::endl;
		}
#endif

	    samples.push_back(smpl);
	}
    }
    
    return count;

}


int savePCA(FileStorage &fs, const PCA& myPCA, const double keep_pca, int max = 0)
{

    uint keepN = 0;
    
    Scalar total = sum(myPCA.eigenvalues);
    const Mat& diag = myPCA.eigenvalues;
    
    double pc = 0;

    if(max<=0)
	max = diag.rows;

    for( ; keepN<diag.rows && keepN<max; keepN++){
	pc += diag.at<float>(keepN,0);
	if(pc > keep_pca * total[0]){
	    // 			myPCA.eigenvectors(Range(0,keepN+1), Range::all()).copyTo(eigVectors);
// 			diag(Range(0,keepN+1), Range::all()).copyTo(eigValues);
	    break;
	}
    }

    if(keepN == (max) && pc < keep_pca*total[0]){
	std::cout << "Maximum number of components reached ("<<keepN<<"). Representation kept: "
		  << pc/total[0]<<std::endl;
    }


    fs << "PCA" << "{pca";//open struct (NODE_MAP)
    fs << "mean"<< myPCA.mean;
    fs << "eigenvectors" << myPCA.eigenvectors(Range(0,keepN+1), Range::all());
    fs << "eigenvalues" << myPCA.eigenvalues(Range(0,keepN+1), Range::all());
    fs << "}";

    std::cout << keepN+1 <<  " vectors written"<< std::endl;

    return 0;
}



int main(int argc, char *argv[])
{


    Mat discard;
    cliparams_t params;
    std::vector<std::vector<Mat> > samples;

    if( argc < 2){
	std::cerr << "Not enough arguments! Use -h for help" << std::endl;    
	return -1;
    }
    
    if(parse_cli(argc, argv, params))
	return 1;
    
 

    //if there is a discard matrix, try to read it
    if(params.discard.length()){
	discard = (CvMat*)cvLoad(params.discard.c_str());	
    }
    
    //load the samples. If the model is holistic, we only want complete samples
    Mat count = loadSamples(params.inputs, samples, discard, !params.per_patch);
    
    //get the patch size
    Size sz;
    int type = -1;
    for(uint k=0;k<samples.at(0).size();k++){
	if(samples.at(0).at(k).data){
	    sz = samples.at(0).at(k).size();
	    type = samples.at(0).at(k).type();
	    break;
	}
    }

    if(type<0){
	std::cerr << "Error: first patch does not have any sample"<<std::endl;
	return -1;
    }

    //if the base is independent for each patch
    //make all samples for each patch into a NxD matrix
    //apply PCA.

    if(params.per_patch){

	std::cout << "Saving in "<< params.output << std::endl;
	FileStorage fs (params.output, FileStorage::WRITE);
	if(!fs.isOpened()){
	    std::cerr << "Error opening file."<<std::endl;
	    return -1;
	}
	fs<<"AppearanceBase"<<"{AppearanceBase";

	fs<<"per_patch_base"<<1;
	fs<<"patch_count"<<(int)samples.at(0).size();
	

	for(uint ai=0; ai< samples.at(0).size();ai++){
	    Mat data(count.at<uint>(ai,0), sz.width * sz.height, type);
	    uint w = 0;
	    for(uint k=0; k<data.rows; k++){
		if(samples.at(k).at(ai).data){
		    Mat r = samples.at(k).at(ai);
		    if(params.normalize){
			normalizar(r,r);
		    }
		    r = r.reshape(0,1);
		    Mat m = data.row(w);
		    r.copyTo(m);
		    w++;
		}
	    }
	    
	    std::cout << "PCA, patch "<< ai<<", on "<< data.rows <<" samples"<<std::endl;
	    PCA myPCA(data, Mat(), CV_PCA_DATA_AS_ROW);


	    //save PCA	    
	    std::stringstream fname;
	    fname <<"patch_"<<std::setfill('0')<<std::setw(3) << ai;
	    fs<< fname.str() <<"{patch_data";

	    savePCA(fs, myPCA, params.keep_pca, params.max_components);       
	    fs << "patch_size" << sz;
	    fs << "patch_count" << 1;
	    fs << "normalized" << (params.normalize ? 1 : 0);
	    fs <<"}";	    


#ifdef DEBUG
	
	Mat projs (data.rows, myPCA.eigenvalues.rows,CV_32F);
	for(uint ai=0; ai<data.rows;ai++){
	    Mat p = myPCA.project(data.row(ai));
	    Mat r = projs.row(ai);
	    p.copyTo(r);
	}	

	Mat ttc;
	reduce (projs, ttc, 0, CV_REDUCE_SUM);
	ttc = ttc.colRange(0,min(10, ttc.cols));
	
	std::cout << ttc << std::endl<<ttc.rows<<std::endl;

#endif
	}

	fs<<"}";//close AppearanceBase

    }
    else{

	Mat data (samples.size(), sz.width * sz.height * count.rows, type);
	Mat sample (sz.height * count.rows, sz.width, type);

	for(uint ai=0; ai < samples.size(); ai++){
	    for(uint k=0; k<samples.at(ai).size();k++){
		Mat m = sample(Range(sz.height*k, sz.height * (k+1)), Range::all());
		if(params.normalize){
		    normalizar(samples.at(ai).at(k), m);
		}
		else
		    samples.at(ai).at(k).copyTo(m);

	    }
	    Mat rd = data.row(ai);
	    Mat r = sample.reshape(0,1);
	    r.copyTo(rd);

	}
	
	std::cout << "PCA, global on "<< data.rows <<" samples"<<std::endl;
	PCA myPCA(data, Mat(), CV_PCA_DATA_AS_ROW);

	std::cout << "Saving in "<< params.output << std::endl;
	FileStorage fs (params.output, FileStorage::WRITE);
	if(!fs.isOpened()){
	    std::cerr << "Error opening file."<<std::endl;
	    return -1;
	}

	fs<<"AppearanceBase"<<"{AppearanceBase";

	fs << "per_patch_base"<<0;
	fs << "patch_count" << (int)samples.at(0).size();

	savePCA(fs, myPCA, params.keep_pca, params.max_components);       
	fs << "patch_size" << sz;
	fs << "normalized" << (params.normalize ? 1 : 0);

	fs<<"}";

#ifdef DEBUG
	
	Mat projs (data.rows, myPCA.eigenvalues.rows,CV_32F);
	for(uint ai=0; ai<data.rows;ai++){
	    Mat p = myPCA.project(data.row(ai));
	    Mat r = projs.row(ai);
	    p.copyTo(r);
	}	

	Mat ttc;
	reduce (projs, ttc, 0, CV_REDUCE_SUM);
	ttc = ttc.colRange(0,min(10,ttc.cols));
	std::cout << ttc << std::endl;

#endif

    }
    
    
    return 0;
}
