// Create a shape deformation vector base using PCA, from a set of
// samples.

/* modelBuildingTools - Software to create texture and shape models
   for Active Models

    Jesus Nuevo-Chiquero, University of Alcal, Spain. 2010
    jesus.nuevochiquero@gmail.com

    This file is part of 'modelBuildingTools'.

    modelBuildingTools is free software: you can redistribute it
    and/or modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    modelBuildingTools is distributed in the hope that it will be
    useful, but WITHOUT ANY WARRANTY; without even the implied
    warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with modelBuildingTools.  If not, see
    <http://www.gnu.org/licenses/>.

*/
#include <stdio.h>
#include <time.h>

#include <sys/stat.h>
#include <sys/types.h>

#include <vector>
#include <stdlib.h>
#include <algorithm>
#include <iostream>
#include <fstream>

#include <stdexcept>

#include <getopt.h>

#include <cxcore.h>
#include <cv.h>
#include <highgui.h>

#include <boost/smart_ptr.hpp>
#include <boost/program_options.hpp>

#include "basic_utils.hh"
#include "procrustes.hh"
#include "bioid_parser.hh"

#include "config.h"

using namespace cv;
namespace po = boost::program_options;

const char *const svn_str = "$Id: build_pdm_base.cpp 679 2009-12-11 15:19:11Z jnuevo $";
const char *const svn_rev = "$Rev: 944 $";

#define DEFAULT_KEEP_PCA 0.95


typedef struct cliparams_t{
    bool use_bioid;
    std::string output;
    double keep_pca;
    int max_components;
    //Mat discard;
    std::string discard;
    std::vector<std::string> inputs;
}cliparams_t;

int parse_cli(int argc, char *argv[], cliparams_t &clarg)
{

    try{

	po::options_description generic("Generic options");
	generic.add_options()
	    ("version,v", "print version string")
	    ("help,h", "produce help message")    
	    ;

	po::options_description config("Configuration");
        config.add_options()
	    ("bioid,b",
	     "The image marks are given in the BioID point format used by D. Cristinacce (default is CvMat XML format)")
	    ("keep,k", po::value<double>(&clarg.keep_pca),
	     "Representation percent to be kept when performing PCA. Default 0.95")
	    ("max_components,m", po::value<int>(&clarg.max_components)->default_value(0),
	     "Maximum number of vectors, regardless if they cover the representation percentage (0 = no limit)")
	    ("discard,d", po::value<std::string>(&clarg.discard)->default_value(""),
	     "A file containing a 1D CvMat, with points to discard from the samples")
	    ;

	po::options_description output("Output");
	output.add_options()
	    ("output,o", po::value<std::string>(), "Output file")
	    ;
	
	// Hidden options, will be allowed both on command line and
        // in config file, but will not be shown to the user.
        po::options_description hidden("Hidden options");
        hidden.add_options()
            ("input-files", po::value< std::vector<std::string> >(&clarg.inputs), "input files")
            ;
	
  	po::positional_options_description p;
        p.add("input-files", -1);

	po::options_description visible;
        visible.add(generic).add(config).add(output);
	

	po::options_description cmdline_options;
        cmdline_options.add(generic).add(config).add(hidden).add(output);

        po::variables_map vm;
        store(po::command_line_parser(argc, argv).
              options(cmdline_options).positional(p).run(), vm);
	
        notify(vm);
	
	if(vm.count("version")){
	    std::cout << "Version: "<<modelBuildingTools_VERSION_MAJOR<<"."
		      <<modelBuildingTools_VERSION_MINOR<<"."
		      <<modelBuildingTools_VERSION_PATCH<<std::endl;
	    std::cout << "SVN rev. : " <<svn_rev << std::endl;
	    return 1;
	}

	if (vm.count("help")) {
	    std::cout << "Usage: build_pdm_base [options] input_files"<<std::endl;
	    std::cout << visible << std::endl;
            return 1;
        }

	if(!vm.count("input-files")){
	    std::cout << "No input files"<<std::endl;
	    return 1;
	}

	if(vm.count("bioid"))
	    clarg.use_bioid = true;
	else
	    clarg.use_bioid = false;

	if(!vm.count("keep")){
	    clarg.keep_pca = DEFAULT_KEEP_PCA;
	}

	if(!vm.count("output")){
	    for(uint ai=0; ai<1e5;ai++){
		char file[128];
		sprintf(file, "pdm_base_ouput_%05d.xml", ai);
		if(access (file, F_OK)){
		    clarg.output = file;
		    break;
		}
	    }
	    if(!clarg.output.length()){
		std::cerr << "Output file could not be created, too many files in current folder"<<std::endl;
		exit(1);
	    }
		
	}
	
    }
    catch(std::exception& e)
	{
	    std::cout << e.what() << "\n";
	    return 1;
	}    
    

    if(clarg.keep_pca>1){
	std::cout << "Bad parameter value: value of -k must be in (0,1]"<<std::endl;
	exit(-1);
    }

    return 0;

}


template<typename T>
std::vector<Mat> loadDataPoints(std::vector<std::string> files, bool use_bioid = false, const std::string &discard_str = "")
{

    std::vector<Mat> points;
    BioIDParser bIDparser;

    Mat discard_mask;        

    //check the discard matrix
    if(discard_str.length()){
	Mat mat;
	//get the first file to read the number of points
	if(use_bioid){
	    bIDparser.load(files[0].c_str());
	    mat = (bIDparser.get_points());	    
	}
	else
	    mat = (CvMat*)cvLoad(files[0].c_str());

	Mat discard = (CvMat*)cvLoad(discard_str.c_str());
	if(discard.data){
	    //first iteration, check discard values are ok
	    bool k = checkRange(discard, true, NULL,
			       0, mat.cols);
	    if(!k || mat.cols <= discard.cols){
		std::cerr<<"Bad discard matrix, bad size or index out of range"<<std::endl;
		exit(-1);
	    }
	    //create a mask
	    
	    discard_mask.create(1, mat.cols, CV_8U);
	    discard_mask = Scalar(255);
	    for(int ai=0; ai<discard.cols; ai++){
		discard_mask.at<uchar>(0,discard.at<int>(0,ai)) = 0;
	    }
	}
    }
    
    for(int ai=0; ai<files.size();ai++){
	Mat mat;
		
	if(use_bioid){
	    bIDparser.load(files[ai].c_str());
	    mat = (bIDparser.get_points());	    
	}
	else
	    mat = (CvMat*)cvLoad(files[ai].c_str());
		

	if(discard_mask.data){	       
	    for(int ai=0, bi=0; ai<mat.cols;ai++){
		if(discard_mask.at<uchar>(0,ai)){
		    mat.row(bi) = mat.row(ai);
		    bi++;
		}
	    }	    
	    mat = mat.rowRange(0, countNonZero(discard_mask)).clone();
	}
	
	else{
	    //check there are no negative values.

	    if(!checkRange(mat, true, NULL, 0)){
		//discard
		std::cerr << "Data "<<ai<<" contains negative values. Discarding..."<<std::endl; 
	    }
	}


	Mat matT = Mat_<T>(mat);//convert to type T
		
	points.push_back(matT);
    }
        

    return points;
}



int main(int argc, char *argv[])
{

    std::vector<Mat> samples;
    int dataorder = DATA_AS_ROWS;
    
 //    show_info();
    

    if( argc < 2){
	std::cerr << "Not enough arguments! Use -h for help" << std::endl;    
	return -1;
    }
    
    
    cliparams_t params;
    if(parse_cli(argc, argv, params))
	return 1;

    //load_data
    std::cout << "Loading Data"<<std::endl;

    if(params.inputs.size()<3){
	std::cerr << params.inputs.size()<<" data vectors. You need more data." << std::endl;
	exit(-1);
    }

   samples = loadDataPoints<float>(params.inputs, params.use_bioid, params.discard);

    Size sz = samples.at(0).size();

    if(sz.width>2 && sz.height == 2){
	dataorder = DATA_AS_COLS;
	for(uint ai=0; ai<samples.size();ai++){
	    Mat tr = samples.at(ai).t();
	    samples.at(ai) = tr.clone();
	    
	}
    }
    std::cout << "Procrustes alignment"<<std::endl;
    //make NxD matrix

    for(uint ai=0; ai<samples.size(); ai++){
	
	removeTranslation(samples.at(ai));
	normalize(samples.at(ai), samples.at(ai),1);
    }

    //apply Procrustes
    Mat gpaMean;
    TermCriteria tC = generalizedProcrustes(samples, gpaMean);

    std::cout << "GProcrustesAnalysis: iterations: "<< tC.maxCount <<", eps: "<<tC.epsilon<<std::endl;
    
    //rotate mean and all data so the line linking the first and
    //second point of the mean is horizontal
    
    Mat rotmat(2,2,CV_32F);
    double alpha = atan2(gpaMean.at<float>(1,1) - gpaMean.at<float>(0,1),
			 gpaMean.at<float>(1,0) - gpaMean.at<float>(0,0));

    rotmat.diag() = cos(alpha);
    rotmat.at<float>(0,1) = sin(alpha);
    rotmat.at<float>(1,0) = -sin(alpha);

    gpaMean = gpaMean * rotmat;

    //reshape data to 1-D vectors. apply PCA
    Mat data(samples.size(), samples[0].rows*samples[0].cols, CV_32F);
    
    for(uint ai=0; ai<samples.size();ai++){
	Mat rp, rd;
	
	samples.at(ai) = samples.at(ai) * rotmat;

	if(dataorder == DATA_AS_COLS){
	    rp = samples.at(ai).t();
	    rp = rp.reshape(0,1);
	}
	else
	    rp = samples.at(ai).reshape(0,1);

	rd = data.row(ai);
	rp.copyTo(rd);

   }
   
    std::cout << "PCA..." <<std::endl;

    PCA myPCA(data, Mat(), CV_PCA_DATA_AS_ROW);

    
    //    std::cout << norm(myPCA.mean.reshape(0,sz.height), gpaMean)<<std::endl;
 
    //keep a given number of vectors
    Mat mean;    
    Mat eigValues, eigVectors;

    myPCA.mean.copyTo(mean);
    uint keepN = 0;
    {
	Scalar total = sum(myPCA.eigenvalues);
	Mat diag = myPCA.eigenvalues;

	double pc = 0;
	for(; keepN<diag.rows;keepN++){
	    pc += diag.at<float>(keepN,0);
	    if(pc > params.keep_pca * total[0] && keepN >= params.max_components){
		myPCA.eigenvectors(Range(0,keepN+1), Range::all()).copyTo(eigVectors);
		diag(Range(0,keepN+1), Range::all()).copyTo(eigValues);
		break;
	    }
	}
    }
    

  
    //check the residue of the data vectors on the PCA.
#ifdef DEBUG
    std::cout << "PCA residues check..." <<std::endl;
    {
	for(uint ai=0;ai<data.rows/50;ai++){
	    Mat proj = myPCA.project(data.row(ai*50));
	    proj.colRange(keepN+1, proj.cols) = Scalar(0);
	    Mat reproj = myPCA.backProject(proj);
	    
	    std::cout <<"Norm reproj ("<<keepN+1<<"): " << norm(data.row(ai*50), reproj);
	    
	    proj = myPCA.project(data.row(ai*50));
	    reproj = myPCA.backProject(proj);
	    std::cout<<". Full base: "<<norm(data.row(ai*50), reproj) <<std::endl;	    
	}

	Mat projs (data.rows, myPCA.eigenvalues.rows,CV_32F);
	for(uint ai=0; ai<data.rows;ai++){
	    Mat p = myPCA.project(data.row(ai));
	    if(ai<10){
		Mat pp = p.colRange(0,10);
		std::cout << pp<<std::endl;
	    }
	    Mat r = projs.row(ai);
	    p.copyTo(r);
	}	

	Mat ttc;
	reduce (projs, ttc, 0, CV_REDUCE_SUM);

	std::cout << ttc << std::endl;

    }

   

#endif
    

    //save eigenvectors & evalues.
    
    //    FileStorage fs (params.ofolder,FileStorage::WRITE);
    FileStorage fs (params.output,FileStorage::WRITE);
    if(fs.isOpened()){
	fs << "pdm_model" << "{pdm_model";
	fs << "pdm_mean" << mean;
	fs << "pdm_vectors" << eigVectors; 
	fs << "pdm_eigvals" << eigValues; 
	fs << "}";
    }

    return 0;

}
