//extract patches from a set of images, given the position of the
//landmarks. These samples are then used in 'build_app_base'

/* modelBuildingTools - Software to create texture and shape models
   for Active Models

    Jesus Nuevo-Chiquero, University of Alcal, Spain. 2010
    jesus.nuevochiquero@gmail.com

    This file is part of 'modelBuildingTools'.

    modelBuildingTools is free software: you can redistribute it
    and/or modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    modelBuildingTools is distributed in the hope that it will be
    useful, but WITHOUT ANY WARRANTY; without even the implied
    warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with modelBuildingTools.  If not, see
    <http://www.gnu.org/licenses/>.

*/
#include <stdio.h>
#include <time.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <stddef.h>
#include <dirent.h>

#include <vector>
#include <stdlib.h>
#include <algorithm>
#include <iostream>
#include <fstream>

#include <stdexcept>

#include <getopt.h>

#include <cxcore.h>
#include <cv.h>
#include <highgui.h>

#include <boost/program_options.hpp>


#include "basic_utils.hh"
#include "bioid_parser.hh"

#include "config.h"

using namespace cv;
namespace po = boost::program_options;

const int DEFAULT_PATCH_WIDTH = 13;
const int DEFAULT_PATCH_HEIGHT = DEFAULT_PATCH_WIDTH;

const char *const svn_str = "$Id: extract_patches.cpp 701 2009-12-22 16:36:01Z jnuevo $";
const char *const svn_rev = "$Rev: 944 $";

//typedef enum  {SUCCESS, DISCARD, DISCARD_ALL, SAVE, SAVE_ALL, QUIT, ERROR_WR} extraction_status_t;


typedef struct cliparams_t{
    bool use_bioid;
    std::string ofolder;
    Size sz;
    std::string discard; //Not yet supported
    double dist; //Not yet supported
    std::vector<std::string> inputs; //folders or list of images & marks
}cliparams_t;


int parse_cli(int argc, char *argv[], cliparams_t &clarg)
{

    std::vector<int> size;

  try{

	po::options_description generic("Generic options");
	generic.add_options()
	    ("version,v", "print version string")
	    ("help,h", "produce help message")    
	    ;

	po::options_description config("Configuration");
        config.add_options()
	    ("bioid,b",
	     "The image marks are given in the BioID point format used by D. Cristinacce (default is CvMat XML format)")
	    ("width,w", po::value<int>(&clarg.sz.width)->default_value(DEFAULT_PATCH_WIDTH),
	     "Width of the patches in pixels")
	    ("height,h", po::value<int>(&clarg.sz.height),
	     "Height of the patches. Equals width by default")
	  //   ("discard,d", po::value<std::string>(&clarg.discard)->default_value(""),
// 	     "A file containing a 1D CvMat, with points to discard from the samples")
	    ;

	po::options_description output("Output");
	output.add_options()
	    ("output,o", po::value<std::string>(&clarg.ofolder), "Output folder")
	    ;
	
	// Hidden options, will be allowed both on command line and
        // in config file, but will not be shown to the user.
        po::options_description hidden("Hidden options");
        hidden.add_options()
            ("input-files", po::value< std::vector<std::string> >(&clarg.inputs), "input files")
            ;
	
  	po::positional_options_description p;
        p.add("input-files", -1);

	po::options_description visible;
        visible.add(generic).add(config).add(output);
	

	po::options_description cmdline_options;
        cmdline_options.add(generic).add(config).add(hidden).add(output);

        po::variables_map vm;
        store(po::command_line_parser(argc, argv).
              options(cmdline_options).positional(p).run(), vm);
	
        notify(vm);
	
	if(vm.count("version")){
	    std::cout << "Version: "<<modelBuildingTools_VERSION_MAJOR<<"."
		      <<modelBuildingTools_VERSION_MINOR<<"."
		      <<modelBuildingTools_VERSION_PATCH<<std::endl;
	    std::cout << "SVN rev. : " <<svn_rev << std::endl;
	}

	if (vm.count("help")) {
	    std::cout << "Usage: extract_patches [options] <image_files marks_files | image_folder marks_folder>"<<std::endl;
	    std::cout << visible << std::endl;
	    std::cout << "Note: if two folders are given as arguments,\n" 
		      << "\timages and marks are read sequentially\n"
		      << "\tfrom both folders. Make sure they correspond."<< std::endl;
            return 1;
        }

	if(!vm.count("input-files")){
	    std::cout << "No input files"<<std::endl;
	    return 1;
	}

	if(vm.count("bioid"))
	    clarg.use_bioid = true;
	else
	    clarg.use_bioid = false;

	if(vm.count("width") && !vm.count("height"))
	    clarg.sz.height = clarg.sz.width;

	if(!vm.count("output")){
	    for(uint ai=0; ai<1e5;ai++){
		char file[128];
		sprintf(file, "patches_extracted_%05d", ai);
		if(access (file, F_OK)){
		    clarg.ofolder = file;
		    break;
		}
	    }
	    if(!clarg.ofolder.length()){
		std::cerr << "Output file could not be created, too many files in current folder"<<std::endl;
		exit(1);
	    }
		
	}
	
    }
    catch(std::exception& e)
	{
	    std::cout << e.what() << "\n";
	    return 1;
	}    
    

    return 0;
}


int extract_patches(const Mat& img, Mat& points, std::vector<Mat> &patches,
		    const cliparams_t &params)
{


    unsigned int out = 0;

//     //read and convert to 32F
//     Mat img (imread(imgfile, CV_LOAD_IMAGE_UNCHANGED));
//     img = Mat_<float>(img);
//     img = img * 1./255;

   

#ifdef DEBUG
    static int wait_time = 0;
    Mat dbgimg (img.size(), CV_32F, 3);
    cvtColor(img, dbgimg, CV_GRAY2BGR);
    namedWindow("Extract_DBG", CV_WINDOW_AUTOSIZE);
#endif    


  //   if(params.dist>1){

//     }


    //Smooth the image with a gaussian
    //   GaussianBlur(img, img, Size(3,3),0,0);

    //extract the patches. The patches are centered around the marks
    
    Rect iRect (Point(0,0), img.size());//size of the image
    for(uint ai=0; ai<points.cols;ai++){
	Mat p;

	//check if the patch is completely inside the image.
	Rect r (points.at<float>(0,ai) - params.sz.width/2,
		points.at<float>(1,ai) - params.sz.height/2,
		params.sz.width, params.sz.height);

	Rect maxR = iRect | r;

	if(maxR == iRect)//patch is in the image
	    p = img(r).clone();
	else
	    out++;

	patches.push_back(p);//pushed even if it is empty.

#ifdef DEBUG
	rectangle(dbgimg, r.tl(), r.br(), CV_RGB(0,255,0));
#endif
    }
    
     
#ifdef DEBUG
 
    imshow("Extract_DBG", dbgimg);
    unsigned char key = static_cast<unsigned char> (waitKey(wait_time));
    if(key == '0' || key == ' '){
	wait_time = 0;
    }		
    else if(key == '1'){
	wait_time = 2;
    }
    else if(key == '2'){
	wait_time = 10;
    }
    else if(key == '3'){
	wait_time = 100;
    }
#endif



    return patches.size() - out;


}


int main(int argc, char *argv[])
{

    cliparams_t params;
    std::vector<std::string> images, marks;
    
    if(argc < 2){
	std::cerr << "Not enough arguments! Use -h for help" << std::endl;    
	return -1;
    }

    
    if(parse_cli(argc, argv, params))
	return 1;

    //check if the inputs are folders or individual files

    int read = read_image_mark_files(params.inputs, params.use_bioid, images, marks);

    if(read <=0){
	std::cerr << "Error reading images and mark files"<<std::endl;
	return -1;
    }

    std::cout << images.size() << " images and marks read"<<std::endl;


    //make the list of files

    //for every image and points file:
    //1.- extract the patches.
    //2.- save them to an xml file in ofolder.
    
    BioIDParser bIDparser;

    for(uint ai=0; ai<images.size();ai++){

	Mat points;
	std::vector<Mat> patches;
	char fn[256];


	if(params.use_bioid){
	    bIDparser.load(marks.at(ai).c_str());
	    points = (bIDparser.get_points());	    
	}
	else
	    points = (CvMat*)cvLoad(marks.at(ai).c_str());
	
	points = Mat_<float>(points);
	if(points.cols == 2){
	    points = points.t();//translate. I don't remember why I
	    //started using 2xN instead of the more
	    //usual Nx2
	}

	//     //read and convert to 32F
	Mat img (imread(images.at(ai), CV_LOAD_IMAGE_UNCHANGED));
	img = Mat_<float>(img);
	img = img * 1./255;
	

	int read = extract_patches(img, points, patches, 
				   params);

	std::cout << "Img: " << images.at(ai) << ", marks: "<<marks.at(ai)
		  <<". "<< read <<" patches read, "
		  << patches.size() - read << " out of the image." <<std::endl;
 
	
	if(read > 0){
	    if(access(params.ofolder.c_str(), F_OK)){//create the folder
		int ret = mkdir(params.ofolder.c_str(), 0777);
		if(ret){
		    std::cerr << "Unable to create or access folder "<< params.ofolder.c_str() <<std::endl
			      << "Exiting..."<<std::endl;
		    return errno;		    			
		}
	    }
	    

	    sprintf(fn, "%s/sample_%05d.xml", params.ofolder.c_str(), ai);
	    FileStorage fs(fn, FileStorage::WRITE);
	    int written=0;
	    
	    if(fs.isOpened()){
		for(uint k=0; k<patches.size();k++){
		    sprintf(fn, "patch_%03d", k);
		    if(patches.at(k).data){
			write(fs, fn, patches.at(k));
			written++;
		    }
		}
		//write the number of patches, total and actually written
		fs << "total_patches"<< (int)patches.size();
		fs << "patches_written"<<written;
		fs << "patch_size" << params.sz;

		fs << "points" << points;

	    }
	    else{
		std::cerr << "Error writing sample "<<ai <<std::endl;
	    }
	}
	else{
	    
	}
	
    }
    

    return 0;

}
