//An implementation of the Procrustes algorithm, used to align shapes.

/* modelBuildingTools - Software to create texture and shape models
   for Active Models

    Jesus Nuevo-Chiquero, University of Alcal, Spain. 2010
    jesus.nuevochiquero@gmail.com

    This file is part of 'modelBuildingTools'.

    modelBuildingTools is free software: you can redistribute it
    and/or modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    modelBuildingTools is distributed in the hope that it will be
    useful, but WITHOUT ANY WARRANTY; without even the implied
    warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with modelBuildingTools.  If not, see
    <http://www.gnu.org/licenses/>.

*/

#include  <stdio.h>

#include <vector>
#include <stdlib.h>
#include <algorithm>
#include <iostream>
#include <fstream>
#include <limits>
#include <iterator>

#include <stdexcept>

#include <cxcore.h>
#include <cv.h>
#include <highgui.h>

#include "basic_utils.hh"
#include "procrustes.hh"

using namespace cv;


Mat removeTranslation(Mat& mat, int dataorder)
{

    Mat suma, m; 

    reduce(mat, suma, dataorder, CV_REDUCE_AVG);
    
    m = repeat(suma,  mat.rows,1);

    mat = mat - m;
    
    return suma;
}



void procrustes(const Mat& X_, Mat& Y,  double *dist, Mat* transform)
{


    Mat covmat;
    Mat X (X_);

    Mat cX = removeTranslation(X);
    Mat cY = removeTranslation(Y);
    double nX = norm(X);
    double nY = norm(Y);

    X = X/nX;
    Y = Y/nY;

    covmat = X.t() * Y;

    SVD svd;

    svd(covmat, SVD::MODIFY_A+ SVD::FULL_UV);

    Mat rotation = svd.vt * svd.u.t();

    Scalar scale = sum(svd.w);

    Y = nX*scale[0] * Y * rotation + repeat(cX,Y.rows,1);

    if(dist)
	*dist = 1 - scale.val[0]*scale.val[0];

    if(transform){
	transform->create(rotation.rows,rotation.cols+1, rotation.type());
	transform->colRange(0,rotation.cols) = scale[0]*nX/nY * rotation;
	transform->col(transform->cols-1) = cX - scale[0]*nX/nY*cY*rotation;
    }

    return;
}


TermCriteria generalizedProcrustes(std::vector<Mat>& samples, Mat& mean, 
				   TermCriteria term)
{

    TermCriteria tC;
    int iter = 0;
    double diff = 1e10; //change from previous mean shape to current mean shape
 
    assert(samples.at(0).type() != CV_32F || samples.at(0).type() != CV_64F);

    Mat prevMean;

    //get the first shape as the mean. 
    mean.create(samples.at(0).size(), samples.at(0).type());
    samples.at(0).copyTo(prevMean);

    //align everything with the mean, until any TermCriteria matches
    while( (!(term.type&TermCriteria::MAX_ITER) || iter < term.maxCount) && 
	   (!(term.type&TermCriteria::EPS) || diff > term.epsilon)){
	
	mean = Scalar(0);

	for(uint ai=0; ai<samples.size();ai++){
	    double dist;

	    procrustes(prevMean, samples.at(ai),&dist);

	    normalize(samples.at(ai),samples.at(ai),1);
	    mean += samples.at(ai);
	}

	//mean = mean * (1./(double)samples.size());
	normalize(mean,mean,1);
	
	diff = norm(mean, prevMean);
	mean.copyTo(prevMean);
	
	iter++;
    }
    
    tC.maxCount = iter;
    tC.epsilon = diff;

    if(term.type&TermCriteria::MAX_ITER && iter == term.maxCount){
	tC.type = TermCriteria::MAX_ITER;
    }
    else{
	tC.type = TermCriteria::EPS;
   }
       
    return tC;
}
