/* modelBuildingTools - Software to create texture and shape models
   for Active Models

    Jesus Nuevo-Chiquero, University of Alcal, Spain. 2010
    jesus.nuevochiquero@gmail.com

    This file is part of 'modelBuildingTools'.

    modelBuildingTools is free software: you can redistribute it
    and/or modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    modelBuildingTools is distributed in the hope that it will be
    useful, but WITHOUT ANY WARRANTY; without even the implied
    warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with modelBuildingTools.  If not, see
    <http://www.gnu.org/licenses/>.

*/
#include <stdio.h>
#include <time.h>

#include <sys/stat.h>
#include <sys/types.h>

#include <vector>
#include <stdlib.h>
#include <algorithm>
#include <iostream>
#include <fstream>
#include <iterator>

#include <stdexcept>

#include <getopt.h>

#include <cxcore.h>
#include <cv.h>
#include <highgui.h>

#include <boost/smart_ptr.hpp>


#include "basic_utils.hh"
#include "procrustes.hh"


//test with data from the example of 'procrustes' from MATLAB


void test_procrustes()
{
    
    std::cout << "********* TEST PROCRUSTES() *********"<<std::endl;
    
    Mat X = (Mat_<double>(10,2) << -0.4326, -0.1867,  -1.6656, 0.7258,
	     0.1253,   -0.5883,
	     0.2877,    2.1832,
	     -1.1465,   -0.1364,
	     1.1909,    0.1139,
	     1.1892,    1.0668,
	     -0.0376,    0.0593,
	     0.3273,   -0.0956,
	     0.1746,   -0.8323); //data from MATLAB

 
    Mat Y = (Mat_<double>(10,2) << 1.8257,    2.1206,
	     1.8311,    2.9372,
	     1.8123,    1.8394,
	     3.0984,    2.4568,
	     1.6197,    2.5269,
	     2.3900,    1.5462,
	     2.8219,    1.8113,
	     1.9366,    1.9710,
	     1.9684,    1.8334,
	     1.7118,    1.7085);
    
    Mat Z =  (Mat_<double>(10,2) <<-0.3488,   -0.1671,
	      -1.6568,    0.6609,
	      0.0901,   -0.4709,
	      0.3877,    2.2175,
	      -1.2089,   -0.0908,
	      1.1414,    0.1640,
	      1.1485,    1.1249,
	      0.0032,  -0.1390,
	      0.2564,   -0.2260,
	      0.1999,   -0.7640);

    double dist = 0;

    //    Mat cX = removeTranslation(X);
    //     Mat cY = removeTranslation(Y);
    //     std::cout << "Centro X: " << cX << std::endl;
    //     std::cout << "Centro Y: " << cY << std::endl;
    //     std::cout << "norm(X): "<<norm(X)<< ", norm(Y): "<< norm(Y) << std::endl;


    procrustes(X, Y,  &dist);    
    
    // std::cout << X<<std::endl;


    std::cout 
	<< "Matrix X: "<<X<<std::endl
	<< "Matrix Y (transformed): "<<Y<<std::endl
	<< "distance: "<<dist<<std::endl;
    std::cout << "Distance to MATLAB data: " << norm(Z,Y)<<std::endl;

    std::cout << "********* END TEST PROCRUSTES() *********"<<std::endl;

}

void test_gpa(int nPts = 10, int nSamples = 5)
{
    
    std::vector<Mat> samples;
    Mat seed (nPts,2,CV_64F);
    Mat mean;

    int r = 0;


    std::cout << "********* TEST GENERALIZED PROCRUSTES ANALYSIS() *********"<<std::endl;
  

    randu(seed, Scalar(-10), Scalar(10));    
    
    //  std::cout << "Seed: "<<std::endl << seed <<std::endl;
 
    for(uint ai=0; ai<nSamples; ai++){

	Mat X(seed.size(),seed.type());
	Mat rotmat (2,2,CV_64F);
	double ang = randu<double>();
	
	randn(X, 0, 1);
	X = seed + X;	

	rotmat.diag() = cos(ang);
	rotmat.at<double>(0,1) = sin(ang);
	rotmat.at<double>(1,0) = -sin(ang);

	X = X *rotmat;

 	removeTranslation(X);
	normalize(X,X,1);

	samples.push_back(X);
    }

    TermCriteria tC = TermCriteria(TermCriteria::MAX_ITER+TermCriteria::EPS, 20, 1e-5 );
    tC = generalizedProcrustes(samples, mean, tC);

    //  std::cout << "Mean: "<<std::endl << mean <<std::endl;    
    std::cout << "TermCriteria: iterations, epsilon: "<< tC.maxCount << ", "<< tC.epsilon << std::endl;
    double d;
    procrustes(mean, seed, &d);
    std::cout << "Dist L2 between seed and mean: "<<d<<std::endl;
 
    std::cout << "********* END TEST GENERALIZED PROCRUSTES ANALYSIS() *********"<<std::endl;
 

    return;
}


int main(int argc, char *argv[])
{

    //   test_procrustes();

    test_gpa();

    test_gpa(10,10);

    test_gpa(40,10);

    test_gpa(40,100);
    
    test_gpa(100,1000);

   return 0;
}
